#!/bin/bash

#set -e

LD=ld
DKMS_BUILD_DIR="$( cd -- "$(dirname "$0")" > /dev/null 2>&1 ; pwd -P)"
PROJECT_DIR=$DKMS_BUILD_DIR/../..
KALLSYMS=$DKMS_BUILD_DIR/kallsyms.txt
UNDEF_SYMS_FILE=$DKMS_BUILD_DIR/undef_syms.txt

FIND_FILTER=("-not" "-path" "*rogue/*" \
	"-not" "-path" "*innosrvkm/bridge/*" \
	"-not" "-path" "*rgx_nohw/*" \
	"-not" "-path" "*rgx_linux_plato/*" \
	"-not" "-path" "*rgx_linux_tc/*" \
	"-not" "-path" "*plato/*" \
	"-not" "-path" "*rgx_vexpress/*" \
	"-not" "-path" "*vz_example/*")

OPEN_SRCS_SUFFIX=_open_srcs.txt
BIN_EXCLUDE_OBJ_SUFFIX=_bin_exclude_objs.tmp

function prepare_env_and_make() {
	cd $PROJECT_DIR
	make -j16
	make export_config
	sed -i 's/^/export /' $DKMS_BUILD_DIR/modules_config.sh
	source $DKMS_BUILD_DIR/modules_config.sh
	cat /proc/kallsyms > $KALLSYMS

	if [[ -n $CROSS_COMPILE ]]; then
		LD=${CROSS_COMPILE}ld
	fi
}

# 模块是否有闭源的二进制
function module_has_private_objs() {
	local module=$1
	objs_dir=$PROJECT_DIR/${module}/
	bin_exclude_objs=$DKMS_BUILD_DIR/${module}${BIN_EXCLUDE_OBJ_SUFFIX}

	find $objs_dir -name "*.o" | grep -vFf $bin_exclude_objs | wc -l
}

# 生成链接二进制需要排除的.o文件
# 该函数是通用实现,不要轻易修改,如果模块有特殊需求自行实现
function prepare_bin_exclude_objs() {
	local module=$1
	open_srcs=$DKMS_BUILD_DIR/${module}${OPEN_SRCS_SUFFIX}
	bin_exclude_objs=$DKMS_BUILD_DIR/${module}${BIN_EXCLUDE_OBJ_SUFFIX}

	cat $open_srcs | grep "\.c" > $bin_exclude_objs
	sed -i 's/\.c/.o/' $bin_exclude_objs
}

# 遍历闭源模块生成的.o文件，检索.o文件中未定义符号是否有引用内核函数
function scan_module() {
	local module=$1
	objs_dir=$PROJECT_DIR/${module}/
	objs_name=$DKMS_BUILD_DIR/${module}_objs.txt
	bin_exclude_objs=$DKMS_BUILD_DIR/${module}${BIN_EXCLUDE_OBJ_SUFFIX}
	cnt=$(module_has_private_objs $module)

	if [[ $cnt != 0 ]]; then
		find $objs_dir -name "*.o" | grep -vFf $bin_exclude_objs > $objs_name
		for i in $(cat $objs_name)
		do
			nm -u $i > $UNDEF_SYMS_FILE
			sed -i 's/.*U //' $UNDEF_SYMS_FILE
			for j in $(cat $UNDEF_SYMS_FILE)
			do
				if [[ $j != __* ]]; then
					grep -wq $j $KALLSYMS;
					if [[ $? == 0 ]]; then
						echo "$j is included at file $i"
					fi
				fi
			done
			rm $UNDEF_SYMS_FILE
		done
	fi
}

function prepare_innogpu_exclude_objs() {
	open_srcs=$DKMS_BUILD_DIR/innogpu${OPEN_SRCS_SUFFIX}
	bin_exclude_objs=$DKMS_BUILD_DIR/innogpu${BIN_EXCLUDE_OBJ_SUFFIX}

	if [[ $CONFIG_INNO_ION != "y" ]]; then
		sed '/ionn*/d' $open_srcs | grep "\.c" > $bin_exclude_objs
	else
		cat $open_srcs | grep "\.c" > $bin_exclude_objs
	fi

	sed -i 's/\.c/.o/' $bin_exclude_objs
}

function prepare_innodma_exclude_objs() {
	prepare_bin_exclude_objs innodma
}

function prepare_innoion_exclude_objs() {
	prepare_bin_exclude_objs innoion
}

function prepare_innopmbus_exclude_objs() {
	prepare_bin_exclude_objs innopmbus
}

function prepare_innovpu_exclude_objs() {
	prepare_bin_exclude_objs innovpu
}

function prepare_innosr_exclude_objs() {
	prepare_bin_exclude_objs innosr
}

function prepare_innopower_exclude_objs() {
	prepare_bin_exclude_objs innopower
}

function prepare_innoapu_exclude_objs() {
	prepare_bin_exclude_objs innoapu_img_mem
	prepare_bin_exclude_objs innoapu_vha
}

function prepare_innosrvkm_exclude_objs() {
	open_srcs=$DKMS_BUILD_DIR/innosrvkm_open_srcs.txt
	bin_objs=$DKMS_BUILD_DIR/innosrvkm${BIN_EXCLUDE_OBJ_SUFFIX}
	_FILE=${open_srcs}.tmp
	cp $open_srcs $_FILE

	if [[ $PVR_BUILD_DIR = android_container ]]; then
		sed -i '/pvr_sync_ioctl_drm.c/d' $_FILE
	else
		sed -i '/pvr_sync_ioctl_dev.*/d' $_FILE
		sed -i '/ion_sys.*/d' $_FILE
	fi

	if [[ $CONFIG_DRM_INNO_POWER != "y" ]]; then
		sed -i '/pvr_dvfs_device.*/d' $_FILE
	fi

	cat $_FILE | grep "\.c" > $bin_objs

	if [[ $CONFIG_DRM_INNO_DPU = "y" ]]; then
		dpu_open_srcs=$DKMS_BUILD_DIR/innodpu_open_srcs.txt
		cat $dpu_open_srcs | grep "\.c" >> $bin_objs
		echo "innogpu_drm.c" >> $bin_objs

		if [[ $CONFIG_DRM_INNO_AUDIO = "y" ]]; then
			audio_open_srcs=$DKMS_BUILD_DIR/innoaudio_open_srcs.txt
			cat $audio_open_srcs | grep "\.c" >> $bin_objs
		fi

	else
		echo "innogpu_drm_without_dpu.c" >> $bin_objs
	fi

	sed -i 's/\.c/.o/' $bin_objs
}

function scan_module_innogpu() {
	scan_module innogpu
}

function scan_module_innodma() {
	scan_module innodma
}

function scan_module_innoion() {
	scan_module innoion
}

function scan_module_innopmbus() {
	scan_module innopmbus
}

function scan_module_innosrvkm() {
	scan_module innosrvkm
}

function scan_module_innovpu() {
	scan_module innovpu
}

function scan_module_innosr() {
	scan_module innosr
}

function scan_module_innopower() {
	scan_module innopower
}

function scan_module_innoapu() {
	local module=innoapu
	local submodule_array=("img_mem" "vha")

	for submodule in "${submodule_array[@]}"
	do
		objs_dir=$PROJECT_DIR/${module}/${submodule}/
		#bin_name=$TARGET_APU/${module}/${submodule}/${submodule}_part.o_shipped
		objs_name=$DKMS_BUILD_DIR/${module}_objs.txt
		bin_exclude_objs=$DKMS_BUILD_DIR/${module}_${submodule}${BIN_EXCLUDE_OBJ_SUFFIX}

		cnt=$(find $objs_dir -name "*.o" -not -name "${submodule}".o -not -name "${submodule}".mod.o | grep -vFf $bin_exclude_objs | wc -l)

		if [[ $cnt != 0 ]]; then
			find $objs_dir -name "*.o" -not -name "${submodule}".o -not -name "${submodule}".mod.o | grep -vFf $bin_exclude_objs > $objs_name
			for i in $(cat $objs_name)
			do
				nm -u $i > $UNDEF_SYMS_FILE
				sed -i 's/.*U //' $UNDEF_SYMS_FILE
				for j in $(cat $UNDEF_SYMS_FILE)
				do
					if [[ $j != __* ]]; then
						grep -wq $j $KALLSYMS;
						if [[ $? == 0 ]]; then
							echo "$j is included at file $i"
						fi
					fi
				done
				rm $UNDEF_SYMS_FILE
			done
		fi
	done
}

function make_module() {
	local module=$1
	prepare_${module}_exclude_objs
	echo -e "\nscan module $module:"
	scan_module_${module}
}

function make_all_modules() {

	make_module innogpu
	make_module innosrvkm
	make_module innodma

	if [[ $CONFIG_DRM_INNO_VPU = "y" ]]; then
		make_module innovpu
	fi

	if [[ $CONFIG_DRM_INNO_SR = "y" ]]; then
		make_module innosr
	fi

	if [[ $CONFIG_INNO_ION = "y" ]]; then
		make_module innoion
	fi

	if [[ $CONFIG_DRM_INNO_PMBUS = "y" ]]; then
		make_module innopmbus
	fi

	if [[ $CONFIG_DRM_INNO_POWER = "y" ]]; then
		make_module innopower
	fi

	if [[ $CONFIG_DRM_INNO_APU = "y" ]]; then
		make_module innoapu
	fi
}

# 清理
function clean() {
	rm -rf $DKMS_BUILD_DIR/*.tmp
	rm -rf $DKMS_BUILD_DIR/modules_config.sh
	rm -rf $DKMS_BUILD_DIR/*_objs.txt
	rm $KALLSYMS
	cd $PROJECT_DIR
	make clean
}

prepare_env_and_make
make_all_modules
clean
